@@ -1185,7 +1185,7 @@ def _run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
     print(f'\nRunning tests with {num_workers} workers')
 
     # Pack the global state
-    state = (compile_commands, clean_commands, test_commands, install_commands, uninstall_commands, backend, backend_flags, host_c_compiler)
+    state = GlobalState(compile_commands, clean_commands, test_commands, install_commands, uninstall_commands, backend, backend_flags, host_c_compiler)
     executor = ProcessPoolExecutor(max_workers=num_workers)
 
     futures: T.List[RunFutureUnion] = []
@@ -1516,10 +1516,15 @@ if __name__ == '__main__':
         # This fails in some CI environments for unknown reasons.
         num_workers = multiprocessing.cpu_count()
     except Exception as e:
-        print('Could not determine number of CPUs due to the following reason:' + str(e))
+        print('Could not determine number of CPUs due to the following reason:', str(e))
         print('Defaulting to using only two processes')
         num_workers = 2
 
+    if num_workers > 64:
+        # Too much parallelism seems to trigger a potential Python bug:
+        # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=1004107
+        num_workers = 64
+
     parser = argparse.ArgumentParser(description="Run the test suite of Meson.")
     parser.add_argument('extra_args', nargs='*',
                         help='arguments that are passed directly to Meson (remember to have -- before these).')
